"""STACS Github integration constants.

SPDX-License-Identifier: BSD-3-Clause
"""

DEFAULT_API_URI = "https://api.github.com"

# Finding is inside of a regular file.
FILE_COMMENT_TEMPLATE = (
    "### :x: [STACS](https://github.com/stacscan/stacs) Finding\n"
    "STACS has found a potential static token or credential at {location} of "
    "`{filename}` due to _{description}_.\n\n"
    "<details><summary>Finding Sample</summary>\n\n"
    "```\n"
    "...{sample}...\n"
    "```\n\n"
    "</details>\n\n"
    "If this credential is valid it should be immediately revoked, and the cause of "
    "this credential making it into this file investigated.\n\n"
    "If this finding is against a 'fake' credential, such as in a test fixture, this "
    "finding can be suppressed using an ignore list in the root of this repository. A "
    "basic ignore list entry can be found below which may be suitable, otherwise, "
    "please refer to the [STACS documentation](https://docs.stacs.app)\n\n"
    "<details><summary>Example Suppression</summary>\n\n"
    "```json\n{suppression}\n```\n\n"
    "</details>\n\n"
    "<sub>[**V**:{version}, **R**:{rule}, **F**:{fhash}]</sub>"
)

# Finding is nested inside of an archive.
NESTED_COMMENT_TEMPLATE = (
    "### :x: [STACS](https://github.com/stacscan/stacs) Finding\n"
    "STACS has found a potential static token or credential at {location} of "
    "`{filename}` due to _{description}_. Please be aware that this file is inside of "
    "an archive, the full path to the file is:\n\n"
    "```\n{tree}\n```\n\n"
    "<details><summary>Finding Sample</summary>\n\n"
    "```\n"
    "...{sample}...\n"
    "```\n\n"
    "</details>\n\n"
    "If this credential is valid it should be immediately revoked, and the cause of "
    "this credential making it into this file investigated.\n\n"
    "If this finding is against a 'fake' credential, such as in a test fixture, this "
    "finding can be suppressed using an ignore list in the root of this repository. A "
    "basic ignore list entry can be found below which may be suitable, otherwise, "
    "please refer to the [STACS documentation](https://docs.stacs.app)\n\n"
    "<details><summary>Example Suppression</summary>\n\n"
    "```json\n{suppression}\n```\n\n"
    "</details>\n\n"
    "<sub>[**V**:{version}, **R**:{rule}, **F**:{fhash}]</sub>"
)



"""Models used by STACS CI.

SPDX-License-Identifier: BSD-3-Clause
"""

from typing import Any, Dict, List

import jmespath
from stacs.ci.exceptions import NoParentException
from stacs.ci.helpers import normalise_string


class SARIFObject:
    def __init__(self, raw: Dict[str, Any]):
        self._raw = raw

    def by_path(self, path: str, default: Any = None) -> Any:
        """Returns a given field by JMESPath, or the default if not found."""
        candidate = jmespath.search(path, self._raw)

        if candidate is None:
            return default
        else:
            return candidate


class Finding(SARIFObject):
    @property
    def location(self) -> str:
        """Returns a plain-text location slug, preferring line numbers to offsets."""
        if self.line:
            return f"line {self.line}"
        else:
            return f"{self.offset}-bytes"

    @property
    def suppressed(self) -> bool:
        """Indicates whether this finding is suppressed."""
        # Determine if this finding is suppressed.
        candidates = self.by_path("suppressions", [])

        if not candidates:
            return False

        # A finding may be listed as suppressed but with a status that isn't 'accepted',
        # in which case the finding isn't actually suppressed.
        for suppression in candidates:
            if suppression.get("status", str()).lower() != "accepted":
                return False

        return True

    @property
    def rule(self):
        """Returns the rule identifier from the finding."""
        return self.by_path("ruleId")

    @property
    def filepath(self):
        """Returns the path to the file the finding was found in."""
        return self.by_path("locations[0].physicalLocation.artifactLocation.uri")

    @property
    def offset(self):
        """Returns the byte offset of the finding."""
        return int(self.by_path("locations[0].physicalLocation.region.byteOffset"))

    @property
    def artifact(self):
        """Returns the artifact index of the finding."""
        return int(self.by_path("locations[0].physicalLocation.artifactLocation.index"))

    @property
    def line(self):
        """Returns the line number which the finding starts at, if applicable."""
        line = self.by_path("locations[0].physicalLocation.region.startLine")

        if line:
            return int(line)
        else:
            return 0

    @property
    def sample(self):
        """Returns the sample from the finding."""
        # Prefer a 'text' sample, if present.
        candidate = self.by_path(
            "locations[0].physicalLocation.contextRegion.snippet.text"
        )
        if candidate:
            return candidate

        # Otherwise return a 'binary' sample.
        return self.by_path(
            "locations[0].physicalLocation.contextRegion.snippet.binary"
        )


class Rule(SARIFObject):
    @property
    def id(self) -> str:
        """Returns the rule identifier."""
        return self.by_path("id", "Unknown")

    @property
    def description(self) -> str:
        """Returns the rule description."""
        return normalise_string(self.by_path("shortDescription.text", "Unknown"))


class Tool(SARIFObject):
    @property
    def version(self) -> str:
        """Returns the tool version."""
        return self.by_path("driver.version", "Unknown")

    @property
    def rules(self) -> List[Rule]:
        """Return a list of Rule objects."""
        candidates = []

        for candidate in self.by_path("driver.rules", []):
            candidates.append(Rule(candidate))

        return candidates


class Artifact(SARIFObject):
    @property
    def filepath(self) -> str:
        """Returns the file path for the artifact"""
        return self.by_path("location.uri")

    @property
    def parent(self) -> int:
        """Gets the artifact index for the parent of this artifact."""
        candidate = self.by_path("parentIndex", None)

        if candidate is not None:
            return int(candidate)
        else:
            raise NoParentException()


class Run(SARIFObject):
    @property
    def findings(self) -> List[Finding]:
        """Returns a list of Finding objects."""
        candidates = []

        for candidate in self.by_path("results", []):
            candidates.append(Finding(candidate))

        return candidates

    @property
    def tool(self) -> str:
        """Returns a Tool object."""
        return Tool(self.by_path("tool"))

    @property
    def artifacts(self) -> List[Artifact]:
        """Returns a list of Artifact objects."""
        candidates = []

        for candidate in self.by_path("artifacts", []):
            candidates.append(Artifact(candidate))

        return candidates


class SARIF(SARIFObject):
    @property
    def version(self) -> str:
        """Returns the version of SARIF this document conforms to."""
        return self.by_path("version")

    @property
    def runs(self) -> List[Run]:
        """Retuns a list of Run objects."""
        candidates = []

        for candidate in self.by_path("runs", []):
            candidates.append(Run(candidate))

        return candidates



"""STACS Generic Integration.

This integration emits an exit status of non-zero if the provided SARIF file contains
any non-suppresssed findings.
"""

import json
import logging
import os
import sys
from json.decoder import JSONDecodeError

from colorama import Fore, init
from stacs.ci import constants, helpers
from stacs.ci.__about__ import __version__
from stacs.ci.models import SARIF


def main(sarif_file: str, prefix: str = None):
    log = logging.getLogger(__name__)

    # Colorama.
    init()

    # Read in the input SARIF file.
    try:
        with open(os.path.abspath(os.path.expanduser(sarif_file)), "r") as fin:
            sarif = SARIF(json.load(fin))
    except (OSError, JSONDecodeError) as err:
        log.fatal(err)
        sys.exit(1)

    # Find all unsuppressed findings, and track them separately.
    results = {}
    findings = 0

    for run in sarif.runs:
        tool = run.tool
        rules = run.tool.rules
        artifacts = run.artifacts

        for finding in run.findings:
            if finding.suppressed:
                continue

            # Get a rule object using the finding rule id.
            rule = None
            for candidate in rules:
                if finding.rule == candidate.id:
                    rule = candidate

            # Just one finding is enough to fail the build.
            findings += 1

            # Construct a virtual path for handling findings in nested files (archives),
            # and get the parent identifier.
            virtual_path = helpers.generate_virtual_path(finding, artifacts)
            if prefix:
                virtual_path = f"{prefix.rstrip('/')}/{virtual_path}"

            if results.get(virtual_path) is None:
                results[virtual_path] = []

            # Generates all strings for presentation now, rather than later.
            results[virtual_path].append(
                {
                    "tree": helpers.get_file_tree(virtual_path),
                    "path": finding.filepath,
                    "rule": finding.rule,
                    "text": rule.description,
                    "offset": finding.offset,
                    "line": finding.line,
                    "location": finding.location,
                    "sample": finding.sample,
                }
            )

    # Provide a summary.
    print(helpers.banner(tool_version=tool.version, version=__version__))

    if findings == 0:
        print("✨ " + Fore.GREEN + "No unsuppressed findings! Great work! ✨\n")
        sys.exit(0)

    # Render out the findings.
    print(
        f"{Fore.RED}🔥 There were {findings} unsuppressed findings in {len(results)} "
        "files 🔥\n"
    )

    for candidate in results:
        filepath = candidate.split(constants.PATH_SEPARATOR)[0]
        count = len(results[candidate])

        if constants.PATH_SEPARATOR in candidate:
            print(f"{Fore.RED}❌ {count} finding(s) inside of file {filepath} (Nested)")
        else:
            print(f"{Fore.RED}❌ {count} finding(s) inside of file {filepath}")

        for finding in results[candidate]:
            print()
            helpers.printi(f"{Fore.YELLOW}Reason   : {finding['text']}")
            helpers.printi(f"{Fore.YELLOW}Rule Id  : {finding['rule']}")
            helpers.printi(f"{Fore.YELLOW}Location : {finding['location']}\n\n")
            helpers.printi(f"{Fore.YELLOW}Filetree:\n\n")
            helpers.printi(
                finding["tree"],
                prefix=f"    {Fore.RESET}|{Fore.BLUE}",
            )
            print()
            helpers.printi(f"{Fore.YELLOW}Sample:\n\n")
            helpers.printi(
                f"... {finding['sample']} ...",
                prefix=f"    {Fore.RESET}|{Fore.BLUE}",
            )
            print()

        print(f"\n{Fore.RESET}{'-' * 78}\n")

    sys.exit(constants.FINDING_EXIT_CODE)



"""Provides a generic entrypoint for the STACS Github integration. """
import argparse
import logging

from stacs.ci.github import pr

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "STACS Github integration: Annotates Github pull requests with STACS "
            "findings."
        )
    )
    parser.add_argument(
        "sarif",
        help="Path to the SARIF file to process",
    )
    parser.add_argument(
        "--prefix",
        help=(
            "The path to a sub-directory under the repository root the scan was"
            "executed from"
        ),
    )
    arguments = parser.parse_args()

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(process)d - [%(levelname)s] %(message)s",
    )
    log = logging.getLogger(__name__)

    # So far we only support pull-requests, but that may change in future.
    pr.main(sarif_file=arguments.sarif, prefix=arguments.prefix)



"""STACS CI Exceptions.

SPDX-License-Identifier: BSD-3-Clause
"""


class STACSIntegrationException(Exception):
    """The most generic form of exception raised by STACS integrations."""


class InvalidFindingException(STACSIntegrationException):
    """Indicates that a finding appears to be malformed."""


class NoParentException(STACSIntegrationException):
    """Indicates that a finding does not have a parent."""


class ChangeNotInDiffException(STACSIntegrationException):
    """Indicates that a finding does not appear in the current diff."""


class ExternalServiceException(STACSIntegrationException):
    """Indicates that an issue occurred while communicating with an external service."""



"""STACS Github integration.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs.ci.github import client  # noqa: F401
from stacs.ci.github import constants  # noqa: F401
from stacs.ci.github import pr  # noqa: F401



"""Helpers used by multiple STACS CI.

SPDX-License-Identifier: BSD-3-Clause
"""

import hashlib
import json
import re
from typing import List

import colorama
from stacs.ci.constants import PATH_SEPARATOR, PATTERN_FHASH
from stacs.ci.exceptions import NoParentException


def generate_virtual_path(
    finding: "Finding",  # noqa: F821
    artifacts: "List[Artifact]",  # noqa: F821
):
    """Generate a virtual path for an input file."""
    virtual_path = finding.filepath

    try:
        parent = artifacts[finding.artifact].parent

        while True:
            name = artifacts[parent].filepath
            virtual_path = f"{name}{PATH_SEPARATOR}{virtual_path}"

            parent = artifacts[parent].parent
    except NoParentException:
        return virtual_path


def generate_fhash(filepath: str, offset: int, rule: str) -> str:
    """Generates a finding hash for use in de-duplicating comments."""
    return hashlib.sha1(bytes(f"{filepath}.{offset}.{rule}", "utf-8")).hexdigest()


def parse_fhashes(content: List[str]) -> List[str]:
    """Parses finding hashes from a list of text - such as Github comments."""
    fhashes = []

    for text in content:
        fhash = re.search(PATTERN_FHASH, text)
        if fhash:
            fhashes.append(fhash.group(1))

    return fhashes


def normalise_string(text: str) -> str:
    """Return the input string without the proceeding capital and trailing full-stop."""
    candidate = str()

    for index, char in enumerate(list(text)):
        if index == 0:
            candidate += char.lower()
        else:
            candidate += char

    return candidate.rstrip(".")


def get_file_tree(virtual_path: str) -> str:
    """Returns a tree layout to the virtual path."""
    tree = str()
    parts = virtual_path.split(PATH_SEPARATOR)

    for index, part in enumerate(parts):
        # Add some style. Print a package / box before each archive, and a document
        # before the file.
        if (index + 1) == len(parts):
            emoji = "📄"
        else:
            emoji = "📦"

        tree += f"{' ' * (index * 4)}`-- {emoji} {part}\n"

    return tree.rstrip()


def generate_suppression(filepath: str) -> str:
    """Generate an example suppression document for the given file."""
    return json.dumps(
        {
            "include": [],
            "ignore": [
                {
                    "pattern": f"{re.escape(filepath)}$",
                    "reason": "A reason for this suppression",
                }
            ],
        },
        indent=4,
        sort_keys=True,
    )


def printi(string, indent: int = 4, prefix: str = None):
    """Super janky wrapper to print something indented."""
    for line in string.splitlines():
        if prefix:
            print(f"{prefix}", end="")

        print(f"{' ' * indent}" + line)


def banner(version: str, tool_version: str) -> str:
    """Returns a STACS console banner."""
    banner = colorama.Fore.BLUE
    banner += rf"""
    ______________   ___________
   / ___/_  __/   | / ____/ ___/
   \__ \ / / / /| |/ /    \__ \
  ___/ // / / ___ / /___ ___/ /
 /____//_/ /_/  |_\____//____/

       STACS version {tool_version}
 STACS Integration Version {version}
    """
    return banner



"""Processes diff files with Github pull-request semantics.

SPDX-License-Identifier: BSD-3-Clause
"""

import re
from typing import Dict, TypedDict

from stacs.ci.constants import PATTERN_GIT_DIFF_HEADER, PATTERN_GIT_DIFF_HUNK


class Hunk(TypedDict):
    content: str
    offset: int


def parse(raw: str) -> Dict[str, Dict[str, Hunk]]:
    """
    Parse a unified diff into a nested dictionary, keyed by filename and then line
    number of the given hunk.
    """
    changes = {}
    pointer = 0
    diff = raw.splitlines()

    while pointer < len(diff):
        # Check if the line is a 'git diff' file header, and if so, prepare a new entry
        # as this indicates a new file in the diff.
        match = re.match(PATTERN_GIT_DIFF_HEADER, diff[pointer])
        if match:
            line = "0"
            hunks = 0
            filename = match.group(2)
            changes[filename] = {}

        # Check if the line is a hunk start.
        match = re.match(PATTERN_GIT_DIFF_HUNK, diff[pointer])

        if match:
            # Track the 'range start' of this hunk hunk in the destination file.
            line = match.group(2).split(",")[0]

            # If this is the first hunk of this specific file, then track the location
            # as Github review comments are indexed by their location relative to the
            # first hunk.
            if hunks == 0:
                offset = pointer

            hunks += 1

        # Extract the contents of this hunk and track as is. Zero being 'Falsy' works to
        # our advantage here, as line 0 only exists if the file is being removed, in
        # which case we want to skip this block anyway.
        if int(line):
            if not changes[filename].get(line):
                changes[filename][line] = {"content": ""}

                # Offsets are referring to the count of lines after the first hunk
                # header for a given file. This is required when working with Github's
                # API for adding pull-request comments.
                if hunks == 1:
                    changes[filename][line]["offset"] = 1
                else:
                    changes[filename][line]["offset"] = (pointer - offset) + 1

            # Track the differences for this hunk as a blob of text. It doesn't matter
            # whether the real diff contains CRLF (\r\n or just LF (\n) as we just need
            # a way to determine when the line ends.
            if diff[pointer].startswith(("+", "-", " ")):
                changes[filename][line]["content"] += f"{diff[pointer]}\n"

        pointer += 1

    return changes



"""STACS Github pull request integration.

SPDX-License-Identifier: BSD-3-Clause
"""

import json
import logging
import os
import sys
from json.decoder import JSONDecodeError
from typing import Dict, List

from stacs.ci import diff, exceptions, helpers
from stacs.ci.constants import FINDING_EXIT_CODE
from stacs.ci.github.client import Client
from stacs.ci.github.constants import FILE_COMMENT_TEMPLATE, NESTED_COMMENT_TEMPLATE
from stacs.ci.models import SARIF


def validate_environment() -> List[str]:
    """Checks whether any required environment variables are missing."""
    missing = []
    required = [
        "GITHUB_TOKEN",
        "GITHUB_API_URL",
        "GITHUB_REPOSITORY",
        "GITHUB_SHA",
        "GITHUB_REF",
    ]

    for candidate in required:
        if os.environ.get(candidate) is None:
            missing.append(candidate)

    return missing


def position_in_diff(
    filepath: str,
    line: int,
    changes: Dict[str, Dict[int, diff.Hunk]],
) -> int:
    """Determine the position that the comment needs to be added in the diff."""
    position = 0

    # Check that the finding is present in a hunk which is part of this pull-request. A
    # finding may already be in a file modified in this pull-request, but not in one of
    # the locations which was modified. In this case, once again, a regular comment
    # will be added as we cannot annotate part of the file which was not modified by
    # this pull-request.
    if filepath not in changes:
        raise exceptions.ChangeNotInDiffException()

    for start, change in changes[filepath].items():
        # Determine the last line number of this hunk by counting all additions and
        # existing lines, but ignoring removals.
        current = int(start)
        contents = change["content"].splitlines()
        offset = change["offset"]

        for index, candidate in enumerate(contents):
            if candidate.startswith(("+", " ")) and index < len(contents) - 1:
                current += 1

            # Track the line number relative to the first diff hunk of this file - which
            # is the required offset to add a comment via the Github API.
            if line == current:
                if current == 1:
                    position = 1
                else:
                    position = offset + (index + 1)

        # If we already have a matching location, return it.
        if position:
            return position

    raise exceptions.ChangeNotInDiffException()


def main(sarif_file: str, prefix: str = None):
    """STACS Github pull request integration."""
    log = logging.getLogger(__name__)

    # Ensure the environment is as we expect (running in Github actions).
    missing_context = validate_environment()
    if missing_context:
        log.fatal(
            "Required environment variables are missing cannot continue: "
            f"{', '.join(missing_context)}"
        )
        sys.exit(1)

    # Read in the input SARIF file.
    try:
        with open(os.path.abspath(os.path.expanduser(sarif_file)), "r") as fin:
            sarif = SARIF(json.load(fin))
    except (OSError, JSONDecodeError) as err:
        log.fatal(err)
        sys.exit(1)

    # Setup a Github API client.
    github = Client(api=os.environ["GITHUB_API_URL"], token=os.environ["GITHUB_TOKEN"])

    # Fetch and parse the pull-request diff from Github.
    log.info("Attempting to fetch and parse pull-request diff")
    try:
        changes = diff.parse(
            github.get_pull_request_diff(
                repository=os.environ["GITHUB_REPOSITORY"],
                reference=os.environ["GITHUB_REF"],
            )
        )
    except exceptions.ExternalServiceException as err:
        log.fatal(err)
        sys.exit(1)

    # Request a list of comments and issue comments from Github.
    comments = []
    try:
        log.info("Attempting to get existing pull-request comments")
        comments.extend(
            github.get_issue_comments(
                repository=os.environ["GITHUB_REPOSITORY"],
                reference=os.environ["GITHUB_REF"],
            )
        )

        log.info("Attempting to get existing pull-request review comments")
        comments.extend(
            github.get_pull_request_review_comments(
                repository=os.environ["GITHUB_REPOSITORY"],
                reference=os.environ["GITHUB_REF"],
            )
        )
    except exceptions.ExternalServiceException as err:
        log.fatal(err)
        sys.exit(1)

    # Parse finding hashes (fhahes) from existing review and issue comments.
    fhashes = helpers.parse_fhashes(comments)

    # Track unsupressed findings so that we can exit appropriately.
    unsuppressed = 0

    # Roll over the findings and add comments where required.
    for run in sarif.runs:
        rules = run.tool.rules
        artifacts = run.artifacts

        for finding in run.findings:
            # Construct a virtual path for handling findings in nested files (archives),
            # and get the parent identifier.
            filepath = finding.filepath
            virtual_path = helpers.generate_virtual_path(finding, artifacts)

            if prefix:
                filepath = f"{prefix.rstrip('/')}/{filepath}"
                virtual_path = f"{prefix.rstrip('/')}/{virtual_path}"

            # Skip the finding if it's suppressed, otherwise track it.
            if finding.suppressed:
                log.info(f"Skipping suppressed {finding.rule} finding in {filepath}")
                continue
            else:
                unsuppressed += 1

            try:
                parent = artifacts[finding.artifact].parent
            except exceptions.NoParentException:
                parent = None

            # Get a rule object using the finding rule id.
            rule = None
            for candidate in rules:
                if finding.rule == candidate.id:
                    rule = candidate

            # Determine if this finding already has a comment using the finding hash.
            fhash = helpers.generate_fhash(virtual_path, finding.offset, finding.rule)
            if fhash in fhashes:
                log.info(
                    f"Found existing comment for {finding.rule} finding in {filepath}, "
                    f"skipping (FHASH:{fhash})"
                )
                continue

            # Only calculate the position in the diff if the file is both text, and is
            # directly in the repository (not inside an archive). We can check if the
            # finding has a line number to quickly check if the finding is inside a
            # binary.
            if finding.line and parent is None:
                # Only add a review comment if the file is present in the review.
                try:
                    position = position_in_diff(filepath, finding.line, changes)

                    log.info(
                        f"Attempting to add review comment for {finding.rule} finding "
                        f"in regular file at {filepath}"
                    )
                    github.add_pull_request_review_comment(
                        repository=os.environ["GITHUB_REPOSITORY"],
                        reference=os.environ["GITHUB_REF"],
                        comment=FILE_COMMENT_TEMPLATE.format(
                            location=finding.location,
                            sample=finding.sample,
                            filename=filepath,
                            rule=rule.id,
                            description=rule.description,
                            fhash=fhash,
                            version=run.tool.version,
                            suppression=helpers.generate_suppression(filepath),
                        ),
                        filepath=filepath,
                        position=position,
                        commit=os.environ["GITHUB_SHA"],
                    )
                    continue
                except exceptions.ChangeNotInDiffException:
                    # This case will be handled later.
                    pass

            # Check if the finding is inside of an archive, and if so, generate a file
            # tree for easier location.
            if parent:
                log.info(
                    f"Attempting to add pull-request comment for {finding.rule} "
                    f"finding in nested file at {filepath}"
                )
                github.add_issue_comment(
                    repository=os.environ["GITHUB_REPOSITORY"],
                    reference=os.environ["GITHUB_REF"],
                    comment=NESTED_COMMENT_TEMPLATE.format(
                        location=finding.location,
                        sample=finding.sample,
                        filename=filepath,
                        tree=helpers.get_file_tree(virtual_path=virtual_path),
                        rule=rule.id,
                        description=rule.description,
                        fhash=fhash,
                        version=run.tool.version,
                        suppression=helpers.generate_suppression(filepath),
                    ),
                )
                continue

            # Otherwise, the finding is likely in a binary directly in the repository,
            # or a text file already in the target branch.
            log.info(
                f"Attempting to add pull-request comment for {finding.rule} "
                f"finding in file at {filepath}"
            )
            github.add_issue_comment(
                repository=os.environ["GITHUB_REPOSITORY"],
                reference=os.environ["GITHUB_REF"],
                comment=FILE_COMMENT_TEMPLATE.format(
                    location=finding.location,
                    sample=finding.sample,
                    filename=filepath,
                    rule=rule.id,
                    description=rule.description,
                    fhash=fhash,
                    version=run.tool.version,
                    suppression=helpers.generate_suppression(filepath),
                ),
            )

    if unsuppressed > 0:
        sys.exit(FINDING_EXIT_CODE)
    else:
        sys.exit(0)



"""Provides a light-weight Github API client that does only what we need.

SPDX-License-Identifier: BSD-3-Clause
"""

from typing import Dict, List

import requests
from stacs.ci import exceptions
from stacs.ci.github.constants import DEFAULT_API_URI


class Client:
    """Provides a light-weight Github API client that does only what we need."""

    def __init__(self, api: str = DEFAULT_API_URI, token: str = None):
        self.api = api.rstrip("/")
        self.token = token

    def _get(
        self,
        endpoint: str,
        headers: Dict[str, str] = None,
        params: Dict[str, str] = None,
    ) -> requests.Response:
        """Wraps an HTTP GET to bolt on required headers."""

        # An authentication token isn't strictly required, so only add if present.
        request_headers = headers
        if self.token:
            request_headers["Authorization"] = f"token {self.token}"

        # Leave it to the caller to validate the response.
        return requests.get(
            f"{self.api}/{endpoint}",
            headers=request_headers,
            params=params,
        )

    def _post(
        self,
        endpoint: str,
        headers: Dict[str, str] = None,
        json: str = None,
    ) -> requests.Response:
        """Wraps an HTTP POST to bolt on required headers."""

        # An authentication token isn't strictly required, so only add if present.
        request_headers = headers
        if self.token:
            request_headers["Authorization"] = f"token {self.token}"

        # Leave it to the caller to validate the response.
        return requests.post(
            f"{self.api}/{endpoint}",
            json=json,
            headers=request_headers,
        )

    def get_pull_request_diff(self, repository: str, reference: str) -> str:
        """Fetches and returns the pull-request diff from Github."""
        pull = reference.split("/")[-2]

        try:
            response = self._get(
                f"repos/{repository}/pulls/{pull}",
                headers={"Accept": "application/vnd.github.v3.diff"},
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise exceptions.ExternalServiceException(
                f"An error occurred fetching the diff for this pull-request: {err}"
            )

        return response.text

    def get_pull_request_review_comments(
        self,
        repository: str,
        reference: str,
    ) -> List[str]:
        """Fetches and returns a list of pull-request review comments from Github."""
        pull = reference.split("/")[-2]
        page = 1
        per_page = 100

        # We only care about the body, so we'll just return a list of text.
        comments = []

        while True:
            try:
                response = self._get(
                    f"repos/{repository}/pulls/{pull}/comments",
                    headers={"Accept": "application/vnd.github.v3+json"},
                    params={
                        "page": page,
                        "per_page": per_page,
                    },
                )
                response.raise_for_status()
            except requests.exceptions.RequestException as err:
                raise exceptions.ExternalServiceException(
                    f"An error occurred fetching the diff for this pull-request: {err}"
                )

            # Track the comment.
            for comment in response.json():
                comments.append(comment.get("body"))

            # Check if there are more records to fetch.
            if len(response.json()) != per_page:
                break

        return comments

    def add_pull_request_review_comment(
        self,
        repository: str,
        reference: str,
        comment: str,
        commit: str,
        filepath: str,
        position: int,
    ):
        """Adds the pull-request review comment to the provided line number."""
        pull = reference.split("/")[-2]

        try:
            response = self._post(
                f"repos/{repository}/pulls/{pull}/comments",
                headers={"Accept": "application/vnd.github.v3+json"},
                json={
                    "body": comment,
                    "commit_id": commit,
                    "path": filepath,
                    "position": position,
                },
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise exceptions.ExternalServiceException(
                f"An error occurred adding a review comment to this pull-request: {err}"
            )

    def get_issue_comments(
        self,
        repository: str,
        reference: str,
    ) -> List[str]:
        """Fetches and returns a list of issue comments from Github."""
        pull = reference.split("/")[-2]
        page = 1
        per_page = 100

        # We only care about the body, so we'll just return a list of text.
        comments = []

        while True:
            try:
                response = self._get(
                    f"repos/{repository}/issues/{pull}/comments",
                    headers={"Accept": "application/vnd.github.v3+json"},
                    params={
                        "page": page,
                        "per_page": per_page,
                    },
                )
                response.raise_for_status()
            except requests.exceptions.RequestException as err:
                raise exceptions.ExternalServiceException(
                    f"An error occurred fetching the diff for this pull-request: {err}"
                )

            # Track the comment.
            for comment in response.json():
                comments.append(comment.get("body"))

            # Check if there are more records to fetch.
            if len(response.json()) != per_page:
                break

        return comments

    def add_issue_comment(self, repository: str, reference: str, comment: str):
        """Adds a comment to a Github issue."""
        pull = reference.split("/")[-2]

        try:
            response = self._post(
                f"repos/{repository}/issues/{pull}/comments",
                headers={"Accept": "application/vnd.github.v3+json"},
                json={"body": comment},
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as err:
            raise exceptions.ExternalServiceException(
                f"An error occurred adding a comment to this pull-request: {err}"
            )



"""Provides a generic entrypoint for the STACS Github integration. """

import argparse
import logging

from stacs.ci.generic import fail

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "STACS CI generic integration: Exits with non-zero status if any "
            "unsuppressed findings are present in the input SARIF document."
        )
    )
    parser.add_argument(
        "sarif",
        help="Path to the SARIF file to process",
    )
    parser.add_argument(
        "--prefix",
        help=(
            "The path to a sub-directory under the repository root the scan was"
            "executed from"
        ),
    )
    arguments = parser.parse_args()

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(process)d - [%(levelname)s] %(message)s",
    )
    log = logging.getLogger(__name__)

    fail.main(sarif_file=arguments.sarif, prefix=arguments.prefix)



"""STACS Integrations Setup.

SPDX-License-Identifier: BSD-3-Clause
"""

import os

from setuptools import find_namespace_packages, setup

# Explicitly pull in the contents of our package's __about__ file due to constraints on
# the previous STACS structure.
__uri__ = None
__title__ = None
__version__ = None

about_file = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "stacs/ci/__about__.py",
)
exec(open(about_file).read())

# Defer to setupmeta.
setup(
    name=__title__,
    setup_requires="setupmeta",
    description="Static Token And Credential Scanner CI Integrations.",
    packages=find_namespace_packages(include=["stacs.*"]),
    url=__uri__,
    version=__version__,
    extras_require={
        "development": [
            "tox",
            "black",
            "flake8",
            "isort",
        ]
    },
)



"""Static Token And Credential Scanner CI integrations.

SPDX-License-Identifier: BSD-3-Clause
"""

from stacs.ci import __about__  # noqa:F401
from stacs.ci import constants  # noqa:F401
from stacs.ci import diff  # noqa:F401
from stacs.ci import exceptions  # noqa:F401
from stacs.ci import generic  # noqa:F401
from stacs.ci import github  # noqa:F401
from stacs.ci import models  # noqa:F401



"""Define constants commonly used throughout STACS CI.

SPDX-License-Identifier: BSD-3-Clause
"""

PATTERN_GIT_DIFF_HEADER = r"^diff\s+.*?a/(.*?)\s+b/(.*)$"
PATTERN_GIT_DIFF_HUNK = r"@@\s+-([0-9,]+)\s+\+([0-9,]+)\s+.*@@"

PATH_SEPARATOR = "!"

# Used to extract hashes of existing comments / bodies of text for deduplication.
PATTERN_FHASH = r"\s+\*\*F\*\*:([a-f0-9]{40})]"

# The exit code to use when there is an unsupressed finding.
FINDING_EXIT_CODE = 100
